#!/usr/bin/env python
# -*- coding: utf-8 -*-

import math

from collections import Counter
from typing import Any, Callable, Optional, Union

import torch

from torch import Tensor


BLEU_COCO_OPTIONS = ("shortest", "average", "closest")


def coco_bleu(
    candidates: list[str],
    mult_references: list[list[str]],
    return_all_scores: bool = True,
    n: int = 4,
    option: str = "closest",
    verbose: int = 0,
    tokenizer: Callable[[str], list[str]] = str.split,
    return_1_to_n: bool = False,
) -> Union[tuple[dict[str, Tensor], dict[str, Tensor]], Tensor]:
    """BiLingual Evaluation Understudy function.

    Paper: https://www.aclweb.org/anthology/P02-1040.pdf

    :param candidates: The list of sentences to evaluate.
    :param mult_references: The list of list of sentences used as target.
    :param return_all_scores: If True, returns a tuple containing the globals and locals scores.
        Otherwise returns a scalar tensor containing the main global score.
        defaults to True.
    :param n: Maximal number of n-grams taken into account. defaults to 4.
    :param option: Corpus reference length mode. Can be "shortest", "average" or "closest". defaults to "closest".
    :param verbose: The verbose level. defaults to 0.
    :param tokenizer: The fast tokenizer used to split sentences into words. defaults to str.split.
    :param return_1_to_no: If True, returns the n-grams results from 1 to n.
        Otherwise return the n-grams scores.
        defauts to False.
    :returns: A tuple of globals and locals scores or a scalar tensor with the main global score.
    """
    cooked_cands, cooked_mrefs = _coco_bleu_update(
        candidates,
        mult_references,
        n,
        tokenizer,
        [],
        [],
    )
    return _coco_bleu_compute(
        cooked_cands,
        cooked_mrefs,
        return_all_scores,
        n,
        option,
        verbose,
        return_1_to_n,
    )


def _coco_bleu_update(
    candidates: list[str],
    mult_references: list[list[str]],
    n: int,
    tokenizer: Callable[[str], list[str]],
    prev_cooked_cands: list,
    prev_cooked_mrefs: list,
) -> tuple[list, list[tuple]]:
    if len(candidates) != len(mult_references):
        raise ValueError(
            f"Invalid number of candidates and references. (found {len(candidates)=} != {len(mult_references)=})"
        )
    new_cooked_mrefs = [
        __cook_references(refs, None, n, tokenizer) for refs in mult_references
    ]
    new_cooked_cands = [
        __cook_candidate(cand, cooked_mrefs_i, None, n, tokenizer)
        for cand, cooked_mrefs_i in zip(candidates, new_cooked_mrefs)
    ]
    prev_cooked_cands += new_cooked_cands
    prev_cooked_mrefs += new_cooked_mrefs
    return prev_cooked_cands, prev_cooked_mrefs


def _coco_bleu_compute(
    cooked_cands: list,
    cooked_mrefs: list,
    return_all_scores: bool = True,
    n: int = 4,
    option: str = "closest",
    verbose: int = 0,
    return_1_to_n: bool = False,
) -> Union[Tensor, tuple[dict[str, Tensor], dict[str, Tensor]]]:
    if option not in BLEU_COCO_OPTIONS:
        raise ValueError(
            f"Invalid option {option=}. (expected one of {BLEU_COCO_OPTIONS})"
        )

    score_1_to_n, scores_1_to_n = __compute_bleu_score(
        cooked_cands,
        cooked_mrefs,
        n,
        option=option,
        verbose=verbose,
    )

    # Note: we use f64 because the original implem use numpy which uses f64 precision
    dtype = torch.float64
    score_n = torch.as_tensor(score_1_to_n[-1], dtype=dtype)
    scores_n = torch.as_tensor(scores_1_to_n[-1], dtype=dtype)
    score_1_to_n = torch.as_tensor(score_1_to_n, dtype=dtype)
    scores_1_to_n = torch.as_tensor(scores_1_to_n, dtype=dtype)

    if return_all_scores:
        global_scores = {
            f"bleu_{n}": score_n,
        }
        local_scores = {
            f"bleu_{n}": scores_n,
        }

        if return_1_to_n:
            global_scores[f"bleu_1_to_{n}"] = score_1_to_n
            local_scores[f"bleu_1_to_{n}"] = scores_1_to_n

        return global_scores, local_scores
    else:
        return score_n


def __cook_sentence(
    s: str,
    n: int = 4,
    tokenizer: Callable[[str], list[str]] = str.split,
) -> tuple[int, Counter[tuple[str, ...]]]:
    """Takes a string as input and returns an object that can be given to
    either cook_refs or cook_test. This is optional: cook_refs and cook_test
    can take string arguments as well."""
    words = tokenizer(s)
    counts = Counter()
    for k in range(1, n + 1):
        for i in range(len(words) - k + 1):
            ngram = tuple(words[i : i + k])
            counts[ngram] += 1
    return len(words), counts


def __cook_references(
    refs: list[str],
    eff: Optional[str],
    n: int,
    tokenizer: Callable[[str], list[str]],
) -> tuple[Union[float, list], dict]:  # lhuang: oracle will call with "average"
    """Takes a list of reference sentences for a single segment
    and returns an object that encapsulates everything that BLEU
    needs to know about them."""

    reflen = []
    maxcounts = {}
    for ref in refs:
        rl, counts = __cook_sentence(ref, n, tokenizer)
        reflen.append(rl)
        for (ngram, count) in counts.items():
            maxcounts[ngram] = max(maxcounts.get(ngram, 0), count)

    # Calculate effective reference sentence length.
    if eff == "shortest":
        reflen = min(reflen)
    elif eff == "average":
        reflen = float(sum(reflen)) / len(reflen)

    # lhuang: N.B.: leave reflen computaiton to the very end!!
    # lhuang: N.B.: in case of "closest", keep a list of reflens!! (bad design)
    return (reflen, maxcounts)


def __cook_candidate(
    test: str,
    reflen_refmaxcounts: tuple[Any, dict[tuple[str, ...], int]],
    eff: Optional[None],
    n: int,
    tokenizer: Callable[[str], list[str]],
) -> dict[str, Any]:
    """Takes a test sentence and returns an object that
    encapsulates everything that BLEU needs to know about it."""

    testlen, counts = __cook_sentence(test, n, tokenizer)
    reflen, refmaxcounts = reflen_refmaxcounts  # Replaces the tuple unpacking

    result = {}
    # Calculate effective reference sentence length.
    if eff == "closest":
        result["reflen"] = min((abs(len - testlen), len) for len in reflen)[1]
    else:  # i.e., "average" or "shortest" or None
        result["reflen"] = reflen

    result["testlen"] = testlen
    result["guess"] = [max(0, testlen - k + 1) for k in range(1, n + 1)]
    result["correct"] = [0] * n

    for (ngram, count) in counts.items():
        result["correct"][len(ngram) - 1] += min(refmaxcounts.get(ngram, 0), count)

    return result


def __compute_bleu_score(
    cooked_cands: list,
    cooked_mrefs: list,
    n: int,
    option: Optional[str] = "closest",
    verbose: int = 0,
) -> tuple[list[float], list[list[float]]]:
    SMALL = 1e-9
    TINY = 1e-15  # so that if guess is 0 still return 0
    bleu_list = [[] for _ in range(n)]

    if option is None:
        option = "average" if len(cooked_mrefs) == 1 else "closest"

    global_cands_len = 0
    global_mrefs_len = 0
    totalcomps = {"testlen": 0, "reflen": 0, "guess": [0] * n, "correct": [0] * n}

    # for each sentence
    for comps in cooked_cands:
        testlen = comps["testlen"]
        global_cands_len += testlen

        reflen = __single_reflen(comps["reflen"], option, testlen)

        global_mrefs_len += reflen

        for key in ("guess", "correct"):
            for k in range(n):
                totalcomps[key][k] += comps[key][k]

        # append per audio bleu score
        bleu = 1.0
        for k in range(n):
            bleu *= (float(comps["correct"][k]) + TINY) / (
                float(comps["guess"][k]) + SMALL
            )
            bleu_list[k].append(bleu ** (1.0 / (k + 1)))

        # N.B.: avoid zero division
        ratio = (testlen + TINY) / (reflen + SMALL)
        if ratio < 1:
            for k in range(n):
                bleu_list[k][-1] *= math.exp(1 - 1 / ratio)

        if verbose > 1:
            print(comps, reflen)

    totalcomps["reflen"] = global_mrefs_len
    totalcomps["testlen"] = global_cands_len

    bleus = []
    bleu = 1.0
    for k in range(n):
        bleu *= float(totalcomps["correct"][k] + TINY) / (
            totalcomps["guess"][k] + SMALL
        )
        bleus.append(bleu ** (1.0 / (k + 1)))
    ratio = (global_cands_len + TINY) / (
        global_mrefs_len + SMALL
    )  # N.B.: avoid zero division
    if ratio < 1:
        for k in range(n):
            bleus[k] *= math.exp(1 - 1 / ratio)

    if verbose > 0:
        print(totalcomps)
        print("ratio:", ratio)

    return bleus, bleu_list


def __single_reflen(
    reflens: list[int],
    option: Optional[str] = None,
    testlen: Optional[int] = None,
) -> float:
    if option == "shortest":
        reflen = min(reflens)
    elif option == "average":
        reflen = float(sum(reflens)) / len(reflens)
    elif option == "closest":
        assert testlen is not None
        reflen = min((abs(len - testlen), len) for len in reflens)[1]
    else:
        raise ValueError(
            f"Invalid argument {option=}. (expected one of {BLEU_COCO_OPTIONS})"
        )

    return reflen



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
BASED ON https://github.com/blmoistawinde/fense/
"""

import logging

from typing import Optional, Union

import numpy as np
import torch

from sentence_transformers import SentenceTransformer
from torch import Tensor
from transformers.models.auto.tokenization_auto import AutoTokenizer
from transformers.tokenization_utils_fast import PreTrainedTokenizerFast

from aac_metrics.functional._fense_utils import (
    BERTFlatClassifier,
    infer_preprocess,
    load_pretrain_echecker,
)


pylog = logging.getLogger(__name__)
ERROR_NAMES = (
    "add_tail",
    "repeat_event",
    "repeat_adv",
    "remove_conj",
    "remove_verb",
    "error",
)


def fense(
    candidates: list[str],
    mult_references: list[list[str]],
    return_all_scores: bool = True,
    sbert_model: Union[str, SentenceTransformer] = "paraphrase-TinyBERT-L6-v2",
    echecker: Union[None, str, BERTFlatClassifier] = "echecker_clotho_audiocaps_base",
    echecker_tokenizer: Optional[AutoTokenizer] = None,
    error_threshold: float = 0.9,
    penalty: float = 0.9,
    device: Union[str, torch.device, None] = "cpu",
    batch_size: int = 32,
    verbose: int = 0,
) -> Union[Tensor, tuple[dict[str, Tensor], dict[str, Tensor]]]:
    """Fluency ENhanced Sentence-bert Evaluation (FENSE)

    Paper: https://arxiv.org/abs/2110.04684
    Original implementation: https://github.com/blmoistawinde/fense

    :param candidates: The list of sentences to evaluate.
    :param mult_references: The list of list of sentences used as target.
    :param return_all_scores: If True, returns a tuple containing the globals and locals scores.
        Otherwise returns a scalar tensor containing the main global score.
        defaults to True.
    :param sbert_model: The sentence BERT model used to extract sentence embeddings for cosine-similarity. defaults to "paraphrase-TinyBERT-L6-v2".
    :param echecker: The echecker model used to detect fluency errors.
        Can be "echecker_clotho_audiocaps_base", "echecker_clotho_audiocaps_tiny", "none" or None.
        defaults to "echecker_clotho_audiocaps_base".
    :param echecker_tokenizer: The tokenizer of the echecker model.
        If None and echecker is not None, this value will be inferred with `echecker.model_type`.
        defaults to None.
    :param error_threshold: The threshold used to detect fluency errors for echecker model. defaults to 0.9.
    :param penalty: The penalty coefficient applied. Higher value means to lower the cos-sim scores when an error is detected. defaults to 0.9.
    :param device: The pytorch device used to run the sBERT and echecker models. defaults to "cpu".
    :param batch_size: The batch size of the sBERT and echecker models. defaults to 32.
    :param verbose: The verbose level. defaults to 0.
    :returns: A tuple of globals and locals scores or a scalar tensor with the main global score.
    """

    # Init models
    sbert_model, echecker, echecker_tokenizer = _load_models_and_tokenizer(
        sbert_model, echecker, echecker_tokenizer
    )

    # Encode sents
    rng_ids = [0]
    for refs in mult_references:
        rng_ids.append(rng_ids[-1] + len(refs))
    flat_references = [ref for refs in mult_references for ref in refs]

    cands_embs = _encode_sents_sbert(sbert_model, candidates, batch_size, verbose)
    mrefs_embs = _encode_sents_sbert(sbert_model, flat_references, batch_size, verbose)

    # Compute sBERT similarities
    sbert_cos_sims = [
        (cands_embs[i] @ mrefs_embs[rng_ids[i] : rng_ids[i + 1]].T).mean().item()
        for i in range(len(cands_embs))
    ]
    sbert_cos_sims = np.array(sbert_cos_sims)

    # Compute and apply fluency error detection penalty
    if echecker is not None and echecker_tokenizer is not None:
        probs_dic = _detect_error_sents(
            echecker,
            echecker_tokenizer,  # type: ignore
            candidates,
            batch_size,
            device,
        )
        fluency_errors = (probs_dic["error"] > error_threshold).astype(float)
        probs_dic = {
            f"fense.{k}_prob": torch.from_numpy(v) for k, v in probs_dic.items()
        }
        fense_scores = sbert_cos_sims * (1.0 - penalty * fluency_errors)
    else:
        fluency_errors = None
        probs_dic = {}
        fense_scores = sbert_cos_sims

    # Aggregate and return
    sbert_cos_sim = sbert_cos_sims.mean()
    fense_score = fense_scores.mean()

    sbert_cos_sim = torch.as_tensor(sbert_cos_sim)
    fense_score = torch.as_tensor(fense_score)
    sbert_cos_sims = torch.from_numpy(sbert_cos_sims)
    fense_scores = torch.from_numpy(fense_scores)

    if return_all_scores:
        sents_scores = {
            "fense": fense_scores,
            "sbert.sim": sbert_cos_sims,
        } | probs_dic
        corpus_scores = {
            "fense": fense_score,
            "sbert.sim": sbert_cos_sim,
        }

        if fluency_errors is not None:
            fluency_errors = torch.from_numpy(fluency_errors)
            fluency_error = fluency_errors.mean()
            sents_scores["fluency_error"] = fluency_errors
            corpus_scores["fluency_error"] = fluency_error

        return corpus_scores, sents_scores
    else:
        return fense_score


def _load_models_and_tokenizer(
    sbert_model: Union[str, SentenceTransformer] = "paraphrase-TinyBERT-L6-v2",
    echecker: Union[None, str, BERTFlatClassifier] = "echecker_clotho_audiocaps_base",
    echecker_tokenizer: Optional[AutoTokenizer] = None,
    device: Union[str, torch.device, None] = "cpu",
) -> tuple[SentenceTransformer, Optional[BERTFlatClassifier], Optional[AutoTokenizer]]:
    if isinstance(sbert_model, str):
        sbert_model = SentenceTransformer(sbert_model, device=device)  # type: ignore
    sbert_model.to(device)

    if isinstance(echecker, str):
        if echecker == "none":
            echecker = None
        else:
            echecker = load_pretrain_echecker(echecker, device)

    if echecker_tokenizer is None and echecker is not None:
        echecker_tokenizer = AutoTokenizer.from_pretrained(echecker.model_type)

    for model in (sbert_model, echecker):
        if model is None:
            continue
        for p in model.parameters():
            p.detach_()
        model.eval()

    return sbert_model, echecker, echecker_tokenizer


def _encode_sents_sbert(
    sbert_model: SentenceTransformer,
    sents: list[str],
    batch_size: int = 32,
    verbose: int = 0,
) -> Tensor:
    return sbert_model.encode(
        sents,
        convert_to_tensor=True,
        normalize_embeddings=True,
        batch_size=batch_size,
        show_progress_bar=verbose >= 2,
    )  # type: ignore


def _detect_error_sents(
    echecker: BERTFlatClassifier,
    echecker_tokenizer: PreTrainedTokenizerFast,
    sents: list[str],
    batch_size: int,
    device: Union[str, torch.device, None],
    max_len: int = 64,
) -> dict[str, np.ndarray]:
    if len(sents) <= batch_size:
        batch = infer_preprocess(
            echecker_tokenizer,
            sents,
            max_len=max_len,
            device=device,
            dtype=torch.long,
        )
        logits: Tensor = echecker(**batch)
        assert not logits.requires_grad
        # batch_logits: (bsize, num_classes=6)
        # note: fix error in the original fense code: https://github.com/blmoistawinde/fense/blob/main/fense/evaluator.py#L69
        probs = logits.sigmoid().transpose(0, 1).cpu().numpy()
        probs_dic = dict(zip(ERROR_NAMES, probs))

    else:
        probs_dic = {name: [] for name in ERROR_NAMES}

        for i in range(0, len(sents), batch_size):
            batch = infer_preprocess(
                echecker_tokenizer,
                sents[i : i + batch_size],
                max_len=max_len,
                device=device,
                dtype=torch.long,
            )

            batch_logits: Tensor = echecker(**batch)
            assert not batch_logits.requires_grad
            # batch_logits: (bsize, num_classes=6)
            # classes: add_tail, repeat_event, repeat_adv, remove_conj, remove_verb, error
            probs = batch_logits.sigmoid().cpu().numpy()

            for j, name in enumerate(probs_dic.keys()):
                probs_dic[name].append(probs[:, j])

        probs_dic = {name: np.concatenate(probs) for name, probs in probs_dic.items()}

    return probs_dic



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Any


def is_mono_sents(sents: Any) -> bool:
    """Returns True if input is list[str]."""
    return isinstance(sents, list) and all(isinstance(sent, str) for sent in sents)


def is_mult_sents(mult_sents: Any) -> bool:
    """Returns True if input is list[list[str]]."""
    return (
        isinstance(mult_sents, list)
        and all(isinstance(sents, list) for sents in mult_sents)
        and all(isinstance(sent, str) for sents in mult_sents for sent in sents)
    )


def _check_input(
    candidates: Any,
    mult_references: Any,
) -> None:
    """Returns True candidates and mult_references have a valid type and size."""
    if not is_mono_sents(candidates):
        raise ValueError("Invalid candidates type. (expected list[str])")

    if not is_mult_sents(mult_references):
        raise ValueError("Invalid mult_references type. (expected list[list[str]])")

    same_len = len(candidates) == len(mult_references)
    if not same_len:
        raise ValueError(
            f"Invalid number of candidates ({len(candidates)}) with the number of references ({len(mult_references)})."
        )

    at_least_1_ref_per_cand = all(len(refs) > 0 for refs in mult_references)
    if not at_least_1_ref_per_cand:
        raise ValueError(
            "Invalid number of references per candidate. (found at least 1 empty list of references)"
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging

from typing import Callable, Iterable, Union

from torch import Tensor

from aac_metrics.classes.base import AACMetric
from aac_metrics.classes.bleu import BLEU
from aac_metrics.classes.meteor import METEOR
from aac_metrics.classes.rouge_l import ROUGEL
from aac_metrics.classes.fense import FENSE
from aac_metrics.classes.spider import SPIDEr
from aac_metrics.functional.evaluate import METRICS_SETS, evaluate


logger = logging.getLogger(__name__)


class Evaluate(AACMetric, list[AACMetric]):
    """Evaluate candidates with multiple references with custom metrics.

    For more information, see :func:`~aac_metrics.functional.evaluate.custom_evaluate`.
    """

    full_state_update = False
    higher_is_better = None
    is_differentiable = False

    def __init__(
        self,
        preprocess: bool = True,
        cache_path: str = "$HOME/.cache",
        java_path: str = "java",
        tmp_path: str = "/tmp",
        verbose: int = 0,
        metrics: Union[str, Iterable[AACMetric]] = "aac",
    ) -> None:
        if isinstance(metrics, str):
            metrics = _get_metrics_classes_list(
                metrics,
                True,
                cache_path,
                java_path,
                tmp_path,
                verbose,
            )

        AACMetric.__init__(self)
        list.__init__(self, metrics)
        self._preprocess = preprocess
        self._cache_path = cache_path
        self._java_path = java_path
        self._tmp_path = tmp_path
        self._verbose = verbose

        self._candidates = []
        self._mult_references = []

    def compute(self) -> tuple[dict[str, Tensor], dict[str, Tensor]]:
        return evaluate(
            self._candidates,
            self._mult_references,
            self._preprocess,
            self,
            cache_path=self._cache_path,
            java_path=self._java_path,
            tmp_path=self._tmp_path,
            verbose=self._verbose,
        )

    def reset(self) -> None:
        self._candidates = []
        self._mult_references = []
        return super().reset()

    def update(
        self,
        candidates: list[str],
        mult_references: list[list[str]],
    ) -> None:
        self._candidates += candidates
        self._mult_references += mult_references


class AACEvaluate(Evaluate):
    """Evaluate candidates with multiple references with all Audio Captioning metrics.

    For more information, see :func:`~aac_metrics.functional.evaluate.aac_evaluate`.
    """

    def __init__(
        self,
        preprocess: bool = True,
        cache_path: str = "$HOME/.cache",
        java_path: str = "java",
        tmp_path: str = "/tmp",
        verbose: int = 0,
    ) -> None:
        super().__init__(
            preprocess,
            cache_path,
            java_path,
            tmp_path,
            verbose,
            "aac",
        )


def _get_metrics_classes_list(
    metric_set_name: str,
    return_all_scores: bool = True,
    cache_path: str = "$HOME/.cache",
    java_path: str = "java",
    tmp_path: str = "/tmp",
    verbose: int = 0,
) -> list[AACMetric]:
    metrics_factory = _get_metrics_classes_factory(
        return_all_scores,
        cache_path,
        java_path,
        tmp_path,
        verbose,
    )

    if metric_set_name in METRICS_SETS:
        metrics = [
            factory()
            for metric_name, factory in metrics_factory.items()
            if metric_name in METRICS_SETS[metric_set_name]
        ]
    else:
        raise ValueError(
            f"Invalid argument {metric_set_name=}. (expected one of {tuple(METRICS_SETS.keys())})"
        )

    return metrics


def _get_metrics_classes_factory(
    return_all_scores: bool = True,
    cache_path: str = "$HOME/.cache",
    java_path: str = "java",
    tmp_path: str = "/tmp",
    verbose: int = 0,
) -> dict[str, Callable[[], AACMetric]]:
    return {
        "bleu_1": lambda: BLEU(
            return_all_scores=return_all_scores,
            n=1,
        ),
        "bleu_2": lambda: BLEU(
            return_all_scores=return_all_scores,
            n=2,
        ),
        "bleu_3": lambda: BLEU(
            return_all_scores=return_all_scores,
            n=3,
        ),
        "bleu_4": lambda: BLEU(
            return_all_scores=return_all_scores,
            n=4,
        ),
        "meteor": lambda: METEOR(
            return_all_scores=return_all_scores,
            cache_path=cache_path,
            java_path=java_path,
            verbose=verbose,
        ),
        "rouge_l": lambda: ROUGEL(
            return_all_scores=return_all_scores,
        ),
        # Note: cider_d and spice and computed inside spider metric
        "spider": lambda: SPIDEr(
            return_all_scores=return_all_scores,
            cache_path=cache_path,
            java_path=java_path,
            tmp_path=tmp_path,
            verbose=verbose,
        ),
        "fense": lambda: FENSE(
            return_all_scores=return_all_scores,
            verbose=verbose,
        ),
    }



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import subprocess

from pathlib import Path
from subprocess import CalledProcessError
from typing import Any, Union


def check_metric_inputs(
    candidates: Any,
    mult_references: Any,
) -> None:
    """Raises ValueError if candidates and mult_references does not have a valid type and size."""
    if not is_mono_sents(candidates):
        raise ValueError("Invalid candidates type. (expected list[str])")

    if not is_mult_sents(mult_references):
        raise ValueError("Invalid mult_references type. (expected list[list[str]])")

    same_len = len(candidates) == len(mult_references)
    if not same_len:
        raise ValueError(
            f"Invalid number of candidates ({len(candidates)}) with the number of references ({len(mult_references)})."
        )

    at_least_1_ref_per_cand = all(len(refs) > 0 for refs in mult_references)
    if not at_least_1_ref_per_cand:
        raise ValueError(
            "Invalid number of references per candidate. (found at least 1 empty list of references)"
        )


def check_java_path(java_path: Union[str, Path]) -> bool:
    """Returns True if the java path is valid."""
    if not isinstance(java_path, (str, Path)):
        return False

    try:
        exitcode = subprocess.check_call(
            [java_path, "--version"],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
    except (CalledProcessError, PermissionError, FileNotFoundError):
        exitcode = 1
    return exitcode == 0


def is_mono_sents(sents: Any) -> bool:
    """Returns True if input is list[str]."""
    return isinstance(sents, list) and all(isinstance(sent, str) for sent in sents)


def is_mult_sents(mult_sents: Any) -> bool:
    """Returns True if input is list[list[str]]."""
    return (
        isinstance(mult_sents, list)
        and all(isinstance(sents, list) for sents in mult_sents)
        and all(isinstance(sent, str) for sents in mult_sents for sent in sents)
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from functools import partial
from typing import Any, Callable, Iterable, Union

import tqdm

from torch import Tensor

from aac_metrics.functional.bleu import bleu
from aac_metrics.functional.meteor import meteor
from aac_metrics.functional.rouge_l import rouge_l
from aac_metrics.functional.fense import fense
from aac_metrics.functional.spider import spider
from aac_metrics.utils.tokenization import preprocess_mono_sents, preprocess_mult_sents


METRICS_SETS = {
    "aac": (
        "bleu_1",
        "bleu_2",
        "bleu_3",
        "bleu_4",
        "meteor",
        "rouge_l",
        "cider_d",
        "spice",
        "spider",
    ),
    "all": (
        "bleu_1",
        "bleu_2",
        "bleu_3",
        "bleu_4",
        "meteor",
        "rouge_l",
        "cider_d",
        "spice",
        "spider",
        "fense",
    ),
}


def evaluate(
    candidates: list[str],
    mult_references: list[list[str]],
    preprocess: bool = True,
    metrics: Union[str, Iterable[Callable[[list, list], tuple]]] = "aac",
    cache_path: str = "$HOME/.cache",
    java_path: str = "java",
    tmp_path: str = "/tmp",
    verbose: int = 0,
) -> tuple[dict[str, Tensor], dict[str, Tensor]]:
    """Evaluate candidates with multiple references with custom metrics.

    :param candidates: The list of sentences to evaluate.
    :param mult_references: The list of list of sentences used as target.
    :param preprocess: If True, the candidates and references will be passed as input to the PTB stanford tokenizer before computing metrics.defaults to True.
    :param metrics: The name of the metric list or the explicit list of metrics to compute. defaults to "aac".
    :param cache_path: The path to the external code directory. defaults to "$HOME/.cache".
    :param java_path: The path to the java executable. defaults to "java".
    :param tmp_path: Temporary directory path. defaults to "/tmp".
    :param verbose: The verbose level. defaults to 0.
    :returns: A tuple of globals and locals scores.
    """
    if isinstance(metrics, str):
        metrics = _get_metrics_functions_list(
            metrics,
            return_all_scores=True,
            cache_path=cache_path,
            java_path=java_path,
            tmp_path=tmp_path,
            verbose=verbose,
        )
    else:
        metrics = list(metrics)

    if preprocess:
        candidates = preprocess_mono_sents(
            candidates, cache_path, java_path, tmp_path, verbose
        )
        mult_references = preprocess_mult_sents(
            mult_references, cache_path, java_path, tmp_path, verbose
        )

    pbar = tqdm.tqdm(
        total=len(metrics), disable=verbose < 2, desc="Computing metrics..."
    )

    global_outs = {}
    local_outs = {}

    for metric in metrics:
        name = metric.__class__.__name__
        pbar.set_description(f"Computing {name} metric...")
        global_outs_i, local_outs_i = metric(candidates, mult_references)
        global_outs |= global_outs_i
        local_outs |= local_outs_i
        pbar.update(1)

    pbar.close()
    return global_outs, local_outs


def aac_evaluate(
    candidates: list[str],
    mult_references: list[list[str]],
    preprocess: bool = True,
    cache_path: str = "$HOME/.cache",
    java_path: str = "java",
    tmp_path: str = "/tmp",
    verbose: int = 0,
) -> tuple[dict[str, Tensor], dict[str, Tensor]]:
    """Evaluate candidates with multiple references with all Audio Captioning metrics.

    :param candidates: The list of sentences to evaluate.
    :param mult_references: The list of list of sentences used as target.
    :param preprocess: If True, the candidates and references will be passed as input to the PTB stanford tokenizer before computing metrics.
        defaults to True.
    :param cache_path: The path to the external code directory. defaults to "$HOME/.cache".
    :param java_path: The path to the java executable. defaults to "java".
    :param tmp_path: Temporary directory path. defaults to "/tmp".
    :param verbose: The verbose level. defaults to 0.
    :returns: A tuple of globals and locals scores.
    """
    return evaluate(
        candidates,
        mult_references,
        preprocess,
        "aac",
        cache_path,
        java_path,
        tmp_path,
        verbose,
    )


def _get_metrics_functions_list(
    metric_set_name: str,
    return_all_scores: bool = True,
    cache_path: str = "$HOME/.cache",
    java_path: str = "java",
    tmp_path: str = "/tmp",
    verbose: int = 0,
) -> list[Callable]:
    metrics_factory = _get_metrics_functions_factory(
        return_all_scores,
        cache_path,
        java_path,
        tmp_path,
        verbose,
    )

    if metric_set_name in METRICS_SETS:
        metrics = [
            factory
            for metric_name, factory in metrics_factory.items()
            if metric_name in METRICS_SETS[metric_set_name]
        ]
    else:
        raise ValueError(
            f"Invalid argument {metric_set_name=}. (expected one of {tuple(METRICS_SETS.keys())})"
        )

    return metrics


def _get_metrics_functions_factory(
    return_all_scores: bool = True,
    cache_path: str = "$HOME/.cache",
    java_path: str = "java",
    tmp_path: str = "/tmp",
    verbose: int = 0,
) -> dict[str, Callable[[list[str], list[list[str]]], Any]]:
    return {
        "bleu_1": partial(
            bleu,
            return_all_scores=return_all_scores,
            n=1,
        ),
        "bleu_2": partial(
            bleu,
            return_all_scores=return_all_scores,
            n=2,
        ),
        "bleu_3": partial(
            bleu,
            return_all_scores=return_all_scores,
            n=3,
        ),
        "bleu_4": partial(
            bleu,
            return_all_scores=return_all_scores,
            n=4,
        ),
        "meteor": partial(
            meteor,
            return_all_scores=return_all_scores,
            cache_path=cache_path,
            java_path=java_path,
            verbose=verbose,
        ),
        "rouge_l": partial(
            rouge_l,
            return_all_scores=return_all_scores,
        ),
        # Note: cider_d and spice and computed inside spider metric
        "spider": partial(
            spider,
            return_all_scores=return_all_scores,
            cache_path=cache_path,
            java_path=java_path,
            tmp_path=tmp_path,
            verbose=verbose,
        ),
        "fense": partial(
            fense,
            return_all_scores=return_all_scores,
            verbose=verbose,
        ),
    }



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import subprocess

from pathlib import Path
from subprocess import CalledProcessError
from typing import Union


def _check_java_path(java_path: Union[str, Path]) -> bool:
    """Returns True if the java path is valid."""
    if not isinstance(java_path, (str, Path)):
        return False

    try:
        exitcode = subprocess.check_call(
            [java_path, "--version"],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
    except (CalledProcessError, PermissionError, FileNotFoundError):
        exitcode = 1
    return exitcode == 0



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import os
import os.path as osp
import subprocess
import sys

from argparse import ArgumentParser, Namespace
from subprocess import CalledProcessError

from torch.hub import download_url_to_file

from aac_metrics.classes.fense import FENSE
from aac_metrics.functional.meteor import FNAME_METEOR_JAR
from aac_metrics.functional.spice import FNAME_SPICE_JAR, DNAME_SPICE_CACHE
from aac_metrics.utils.tokenization import FNAME_STANFORD_CORENLP_3_4_1_JAR


logger = logging.getLogger(__name__)


JAR_URLS = {
    "meteor": {
        "url": "https://github.com/tylin/coco-caption/raw/master/pycocoevalcap/meteor/meteor-1.5.jar",
        "fname": "meteor-1.5.jar",
    },
    "meteor_data": {
        "url": "https://github.com/tylin/coco-caption/raw/master/pycocoevalcap/meteor/data/paraphrase-en.gz",
        "fname": osp.join("data", "paraphrase-en.gz"),
    },
    "spice": {
        "url": "https://github.com/tylin/coco-caption/raw/master/pycocoevalcap/spice/spice-1.0.jar",
        "fname": "spice-1.0.jar",
    },
    "stanford_nlp": {
        "url": "https://github.com/tylin/coco-caption/raw/master/pycocoevalcap/tokenizer/stanford-corenlp-3.4.1.jar",
        "fname": "stanford-corenlp-3.4.1.jar",
    },
}


def download(
    cache_path: str = "$HOME/.cache",
    tmp_path: str = "/tmp",
    verbose: int = 0,
) -> None:
    """Download the code needed for SPICE, METEOR and PTB Tokenizer.

    :param cache_path: The path to the external code directory. defaults to "$HOME/.cache".
    :param verbose: The verbose level. defaults to 0.
    """
    cache_path = osp.expandvars(cache_path)
    tmp_path = osp.expandvars(tmp_path)

    os.makedirs(cache_path, exist_ok=True)
    os.makedirs(tmp_path, exist_ok=True)

    # Download JAR file for tokenization
    stanford_nlp_dpath = osp.join(
        cache_path, osp.dirname(FNAME_STANFORD_CORENLP_3_4_1_JAR)
    )
    os.makedirs(stanford_nlp_dpath, exist_ok=True)

    name = "stanford_nlp"
    info = JAR_URLS[name]
    url = info["url"]
    fname = info["fname"]
    fpath = osp.join(stanford_nlp_dpath, fname)
    if not osp.isfile(fpath):
        if verbose >= 1:
            logger.info(
                f"Downloading jar source for '{name}' in directory {stanford_nlp_dpath}."
            )
        download_url_to_file(url, fpath, progress=verbose >= 1)
    else:
        if verbose >= 1:
            logger.info(f"Stanford model file '{name}' is already downloaded.")

    # Download JAR files for METEOR metric
    meteor_dpath = osp.join(cache_path, osp.dirname(FNAME_METEOR_JAR))
    os.makedirs(meteor_dpath, exist_ok=True)

    for name in ("meteor", "meteor_data"):
        info = JAR_URLS[name]
        url = info["url"]
        fname = info["fname"]
        subdir = osp.dirname(fname)
        fpath = osp.join(meteor_dpath, fname)

        if not osp.isfile(fpath):
            if verbose >= 1:
                logger.info(
                    f"Downloading jar source for '{name}' in directory {meteor_dpath}."
                )
            if subdir not in ("", "."):
                os.makedirs(osp.join(meteor_dpath, subdir), exist_ok=True)

            download_url_to_file(
                url,
                fpath,
                progress=verbose >= 1,
            )
        else:
            if verbose >= 1:
                logger.info(f"Meteor file '{name}' is already downloaded.")

    # Download JAR files for SPICE metric
    spice_jar_dpath = osp.join(cache_path, osp.dirname(FNAME_SPICE_JAR))
    os.makedirs(spice_jar_dpath, exist_ok=True)

    spice_cache_path = osp.join(cache_path, DNAME_SPICE_CACHE)
    os.makedirs(spice_cache_path, exist_ok=True)

    script_path = osp.join(osp.dirname(__file__), "..", "..", "install_spice.sh")
    if not osp.isfile(script_path):
        raise FileNotFoundError(f"Cannot find script '{osp.basename(script_path)}'.")

    if verbose >= 1:
        logger.info(
            f"Downloading JAR sources for SPICE metric into '{spice_jar_dpath}'..."
        )

    command = ["bash", script_path, spice_jar_dpath]
    try:
        subprocess.check_call(
            command,
            stdout=None if verbose >= 2 else subprocess.DEVNULL,
            stderr=None if verbose >= 2 else subprocess.DEVNULL,
        )
    except (CalledProcessError, PermissionError) as err:
        logger.error(err)

    # Download models files for FENSE metric
    if verbose >= 1:
        logger.info("Downloading sBert and Bert error detector for FENSE metric...")
    _ = FENSE(device="cpu")


def _get_main_download_args() -> Namespace:
    parser = ArgumentParser(
        description="Download models and external code to evaluate captions."
    )

    parser.add_argument(
        "--cache_path",
        type=str,
        default="$HOME/.cache",
        help="Cache directory path.",
    )
    parser.add_argument(
        "--tmp_path",
        type=str,
        default="/tmp",
        help="Temporary directory path.",
    )
    parser.add_argument("--verbose", type=int, default=1, help="Verbose level.")

    args = parser.parse_args()
    return args


def _main_download() -> None:
    format_ = "[%(asctime)s][%(name)s][%(levelname)s] - %(message)s"
    handler = logging.StreamHandler(sys.stdout)
    handler.setFormatter(logging.Formatter(format_))
    pkg_logger = logging.getLogger("aac_metrics")
    pkg_logger.addHandler(handler)

    args = _get_main_download_args()

    level = logging.INFO if args.verbose <= 1 else logging.DEBUG
    pkg_logger.setLevel(level)

    download(args.cache_path, args.tmp_path, args.verbose)


if __name__ == "__main__":
    _main_download()
